@@ -7,26 +7,26 @@
 from __future__ import absolute_import, division, with_statement
 
 import sys
+from datetime import timedelta
 
 from past.utils import old_div
 from past.builtins import basestring
 
 from franz.miniclient.agjson import encode_json
+from franz.openrdf.repository.attributes import AttributeFilter
 from franz.openrdf.rio.formats import Format
 
-if sys.version_info[0] > 2:
-    from urllib.parse import quote, urlparse
-else:
-    from urllib import quote
-    from urlparse import urlparse
+from six.moves.urllib.parse import quote, urlparse
 
-import inspect, math, threading
+import copy, inspect, math, re, threading
 
 from contextlib import contextmanager
-from .request import *
+from .request import encode, serialize, decode, deserialize, \
+    jsonRequest, nullRequest, urlenc, RequestError
 from ..openrdf.util.contexts import wrap_context
 from ..openrdf.util.strings import to_native_string
 
+from six import python_2_unicode_compatible
 
 def _split_proxy(proxy):
     """
@@ -47,9 +47,11 @@ def _split_proxy(proxy):
 
 class Service(object):
     def __init__(self, url, user=None, password=None, cainfo=None, sslcert=None,
-                 verifyhost=None, verifypeer=None, proxy=None, session=None):
+                 verifyhost=None, verifypeer=None, proxy=None):
         # Note: it is important to save the arguments into fields with identical names.
         # This makes the weird cloning mechanism used by subclasses work.
+        # WARNING: The cloning process mentioned above will create a **shallow** copy
+        # of each argument.
         self.url = url
         self.user = user
         self.password = password
@@ -60,9 +62,39 @@ class Service(object):
         self.verifypeer = verifypeer
         self.proxy = proxy
         self.proxy_type, self.proxy_host, self.proxy_port = _split_proxy(proxy)
-        self.session = session
+        self.session = None  # Will be created lazily
+        self.transaction_settings = None
+        self.user_attributes = None
 
-    def _instance_from_url(self, subclass, url):
+    def getHeaders(self):
+        """
+        Return a dictionary of headers that should be added to every request.
+        Return ``None`` instead of an empty dictionary if there are no such
+        headers.
+        """
+        s = self.transaction_settings
+        values = []
+        if s:
+            if s.distributed_transaction_timeout is not None:
+                values.append(
+                    ('distributedTransactionTimeout',
+                     time_in_seconds(s.distributed_transaction_timeout)))
+            if s.durability is not None:
+                values.append(('durability', s.durability))
+            if s.transaction_latency_count is not None:
+                values.append(('transactionLatencyCount',
+                               s.transaction_latency_count))
+            if s.transaction_latency_timeout is not None:
+                values.append(('transactionLatencyTimeout',
+                               time_in_seconds(s.transaction_latency_timeout)))
+        result = {}
+        if values:
+            result['x-repl-settings'] = ' '.join('%s=%s' % value for value in values)
+        if self.user_attributes:
+            result['x-user-attributes'] = encode_json(self.user_attributes)
+        return result or None
+
+    def _instance_from_url(self, subclass, url=None):
         """
         Create a copy of this service object that will:
            - be an instance of the ``subclass`` class.
@@ -75,11 +107,15 @@ class Service(object):
         kwargs = {}
         for arg_name in inspect.getargspec(subclass.__init__).args:
             if hasattr(self, arg_name):
-                kwargs[arg_name] = getattr(self, arg_name)
+                kwargs[arg_name] = copy.copy(getattr(self, arg_name))
         # Override URL
-        kwargs['url'] = url
+        if url is not None:
+            kwargs['url'] = url
         return subclass(**kwargs)
 
+    def copy(self):
+        return self._instance_from_url(type(self))
+
 
 class Catalog(Service):
     def listRepositories(self):
@@ -359,10 +395,16 @@ class Repository(Service):
         return jsonRequest(self, "GET", "/statements/id", urlenc(id=ids),
                            accept=accept, callback=callback)
 
-    def addStatement(self, subj, pred, obj, context=None):
+    def addStatement(self, subj, pred, obj, context=None, attributes=None):
         """Add a single statement to the repository."""
-        nullRequest(self, "POST", "/statements", encode_json([[subj, pred, obj, context]]),
-                    content_type="application/json")
+        args = {
+            'subj': subj,
+            'pred': pred,
+            'obj': obj,
+            'context': context,
+            'attributes': attributes and encode_json(attributes)
+        }
+        nullRequest(self, "PUT", "/statement?" + urlenc(**args))
 
     def deleteMatchingStatements(self, subj=None, pred=None, obj=None, context=None):
         """Delete all statements matching the constraints from the
@@ -372,21 +414,27 @@ class Repository(Service):
 
     def addStatements(self, quads, commitEvery=None):
         """Add a collection of statements to the repository. Quads
-        should be an array of four-element arrays, where the fourth
-        element, the graph name, may be None."""
-        nullRequest(self, "POST", "/statements?" + urlenc(commit=commitEvery),
+        should be an array of or five four-element arrays, where the fourth
+        element, the graph name, may be None. The fifth element, if present,
+        must be a string with a JSON-encoded dictionary of attribute values.
+        """
+        nullRequest(self, "POST", "/statements?" +
+                    urlenc(commit=commitEvery),
                     encode_json(quads), content_type="application/json")
 
-    def loadData(self, data, rdf_format, base_uri=None, context=None, commit_every=None, content_encoding=None):
+    def loadData(self, data, rdf_format, base_uri=None, context=None,
+                 commit_every=None, content_encoding=None, attributes=None):
         nullRequest(self, "POST",
                     "/statements?" + urlenc(context=context,
                                             baseURI=base_uri,
-                                            commit=commit_every),
+                                            commit=commit_every,
+                                            attributes=attributes and encode_json(attributes)),
                     data,
                     content_type=Format.mime_type_for_format(rdf_format),
                     content_encoding=content_encoding)
 
-    def loadFile(self, file, rdf_format, baseURI=None, context=None, serverSide=False, commitEvery=None, content_encoding=None):
+    def loadFile(self, file, rdf_format, baseURI=None, context=None, serverSide=False,
+                 commitEvery=None, content_encoding=None, attributes=None):
         mime = Format.mime_type_for_format(rdf_format)
 
         if serverSide:
@@ -403,7 +451,8 @@ class Repository(Service):
             file = None
 
         with body_context as body:
-            params = urlenc(file=file, context=context, baseURI=baseURI, commit=commitEvery)
+            params = urlenc(file=file, context=context, baseURI=baseURI, commit=commitEvery,
+                            attributes=attributes and encode_json(attributes))
             nullRequest(self, "POST", "/statements?" + params, body,
                         content_type=mime, content_encoding=content_encoding)
 
@@ -863,3 +912,60 @@ class Repository(Service):
         finally:
             del self._saveFile
             del self._saveAccept
+
+    def setAttributeFilter(self, attribute_filter):
+        if isinstance(attribute_filter, AttributeFilter):
+            attribute_filter = attribute_filter.to_expr()
+        args = urlenc(filter=attribute_filter)
+        return nullRequest(self, 'POST', '/attributes/staticFilter?' + args)
+
+    def getAttributeFilter(self):
+        return jsonRequest(self, 'GET', '/attributes/staticFilter')
+
+    def clearAttributeFilter(self):
+        return nullRequest(self, 'DELETE', '/attributes/staticFilter')
+
+    def getAttributeDefinitions(self):
+        return jsonRequest(self, 'GET', '/attributes/definitions')
+
+    def getAttributeDefinition(self, name):
+        return jsonRequest(
+            self, 'GET', '/attributes/definitions?' + urlenc(name=name))
+
+    def setAttributeDefinition(self, attr_def):
+        args = {
+            'name': attr_def.name,
+            'allowed-values': attr_def.allowed_values,
+            'ordered': attr_def.ordered,
+            'minimum-number': attr_def.minimum_number,
+            'maximum-number': attr_def.maximum_number
+        }
+        return nullRequest(
+            self, 'POST', '/attributes/definitions?' + urlenc(**args))
+
+    def deleteAttributeDefinition(self, name):
+        return jsonRequest(
+            self, 'DELETE', '/attributes/definitions?' + urlenc(name=name))
+
+
+def time_in_seconds(t):
+    """
+    Try to convert an argument to a number of seconds.
+
+    This works on:
+
+        - integers (which are simply returned)
+        - strings (cast to ints, suffixes like s, h, .. are *not* supported)
+        - timedelta objects
+
+    :param t: Time in seconds.
+    :type t: int|str|timedelta
+    :return: Number of seconds.
+    :rtype: int
+    """
+    if t is None:
+        return None
+    if isinstance(t, timedelta):
+        # Can't use total_seconds() in Python 2.6
+        return t.seconds + t.days * 24 * 3600
+    return int(t)
